﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Persistence.Default.Xml;
using HeuristicLab.Persistence.Interfaces;
using System.Text;

namespace HeuristicLab.Persistence.GUI {

  public partial class PersistenceConfigurationForm : Form {

    private readonly Dictionary<string, IFormatter> formatterTable;
    private readonly Dictionary<string, Type> typeNameTable;    

    public PersistenceConfigurationForm() {
      InitializeComponent();
      formatterTable = new Dictionary<string, IFormatter>();
      typeNameTable = new Dictionary<string, Type>();
      decomposerList.Items.Clear();
      foreach ( IDecomposer decomposer in ConfigurationService.Instance.AllDecomposers ) {
        var item = decomposerList.Items.Add(decomposer.GetType().Name);
        item.Checked = true;
        item.Tag = decomposer;
      }

      foreach ( var formats in ConfigurationService.Instance.AllFormatters ) {
        TabPage page = new TabPage(formats.Key.Name) {Tag = formats.Key};
        formatterTabs.TabPages.Add(page);
        DataGridView gridView = new DataGridView {
          Dock = DockStyle.Fill,
          EditMode = DataGridViewEditMode.EditOnEnter,
          AllowUserToAddRows = false,
          AllowUserToDeleteRows = false,
          AllowUserToResizeRows = false,
          Name = "GridView",
        };
        gridView.CellValueChanged += gridView_CellValueChanged;
        page.Controls.Add(gridView);        
        gridView.Columns.Add(new DataGridViewTextBoxColumn {
          Name = "Type", ReadOnly = true,
          AutoSizeMode = DataGridViewAutoSizeColumnMode.Fill
        });
        gridView.Columns.Add(new DataGridViewCheckBoxColumn
          { AutoSizeMode = DataGridViewAutoSizeColumnMode.DisplayedCells });
        gridView.Columns.Add(new DataGridViewComboBoxColumn
          { AutoSizeMode = DataGridViewAutoSizeColumnMode.Fill });                
        var formatterMap = new Dictionary<string, List<string>>();
        foreach (var formatter in formats.Value) {
          string formatterName = formatter.GetType().Name;
          if (formatterTable.ContainsKey(formatterName)) {
            IFormatter otherFormatter = formatterTable[formatterName];
            formatterTable.Remove(formatterName);
            formatterTable.Add(otherFormatter.GetType().FullName, otherFormatter);
            formatterName = formatter.GetType().FullName;
          }
          formatterTable.Add(formatterName, formatter);
          string typeName = formatter.Type.Name;
          if (typeNameTable.ContainsKey(typeName)) {
            Type otherType = typeNameTable[typeName];
            if (otherType != formatter.Type) {
              typeNameTable.Remove(typeName);
              typeNameTable.Add(otherType.FullName, otherType);
              typeName = formatter.Type.FullName;
              typeNameTable.Add(typeName, formatter.Type);
            }
          } else {
            typeNameTable.Add(typeName, formatter.Type);
          }          
        }
        foreach (var formatter in formats.Value) {          
          string formatterName = 
            formatterTable.ContainsKey(formatter.GetType().Name) ?
            formatter.GetType().Name :
            formatter.GetType().FullName;
          string typeName =
            typeNameTable.ContainsKey(formatter.Type.Name) ?
            formatter.Type.Name :
            formatter.Type.FullName;          
          if (!formatterMap.ContainsKey(typeName))
            formatterMap.Add(typeName, new List<string>());
          formatterMap[typeName].Add(formatterName);
        }
        foreach ( var formatterMapping in formatterMap ) {
          var row = gridView.Rows[gridView.Rows.Add()];
          row.Cells[0].Value = formatterMapping.Key;
          row.Cells[0].ToolTipText = formatterMapping.Key;
          row.Cells[1].Value = true;
          var comboBoxCell = (DataGridViewComboBoxCell) row.Cells[2];          
          foreach ( var formatter in formatterMapping.Value ) {
            comboBoxCell.Items.Add(formatter);
          }          
          comboBoxCell.Value = comboBoxCell.Items[0];          
          comboBoxCell.ToolTipText = comboBoxCell.Items[0].ToString();
          if (comboBoxCell.Items.Count == 1) {
            comboBoxCell.ReadOnly = true;
            comboBoxCell.DisplayStyle = DataGridViewComboBoxDisplayStyle.Nothing;
          }     
        }      
      }
    }

    void gridView_CellValueChanged(object sender, DataGridViewCellEventArgs e) {
      UpdatePreview();
    }    

    private void decomposerList_ItemDrag(object sender, ItemDragEventArgs e) {
      decomposerList.DoDragDrop(decomposerList.SelectedItems, DragDropEffects.Move);
    }

    private void decomposerList_DragEnter(object sender, DragEventArgs e) {      
      if ( e.Data.GetDataPresent(typeof(ListView.SelectedListViewItemCollection).FullName)) {
        e.Effect = DragDropEffects.Move;
      } 
    }

    private void decomposerList_DragDrop(object sender, DragEventArgs e) {
      if (decomposerList.SelectedItems.Count == 0) {
        return;
      }
      Point cp = decomposerList.PointToClient(new Point(e.X, e.Y));      
      ListViewItem targetItem = decomposerList.GetItemAt(cp.X, cp.Y);      
      if (targetItem == null)
        return;            
      int targetIndex = targetItem.Index;      
      var selectedItems = new List<ListViewItem>(decomposerList.SelectedItems.Cast<ListViewItem>());
      int i = 0;
      foreach ( ListViewItem dragItem in selectedItems ) {                
        if (targetIndex == dragItem.Index)
          return;
        if (dragItem.Index < targetIndex) {
          decomposerList.Items.Insert(targetIndex + 1, (ListViewItem) dragItem.Clone());
        } else {
          decomposerList.Items.Insert(targetIndex + i, (ListViewItem)dragItem.Clone());          
        }        
        decomposerList.Items.Remove(dragItem);
        i++;
      }
      UpdatePreview();
    }

    private void decomposerList_Resize(object sender, EventArgs e) {
      DecomposersColumn.Width = decomposerList.Width - 4;
    }

    private IEnumerable<IDecomposer> GetDecomposers() {
      List<IDecomposer> decomposers = new List<IDecomposer>();
      foreach ( ListViewItem item in decomposerList.Items ) {
        if (item != null && item.Checked)
          decomposers.Add((IDecomposer) item.Tag);
      }      
      return decomposers;
    }

    private void UpdatePreview() {
      checkbox.Items.Clear();      
      foreach (var decomposer in GetDecomposers())
        checkbox.Items.Add(decomposer.GetType().Name + " (D)");
      IFormat activeFormat = (IFormat) formatterTabs.SelectedTab.Tag;
      if (activeFormat != null) {
        foreach (var formatter in GetCurrentConfiguration(activeFormat).Formatters) {
          checkbox.Items.Add(formatter.GetType().Name + " (F)");
        }
      }
    }

    private void decomposerList_ItemChecked(object sender, ItemCheckedEventArgs e) {
      UpdatePreview();
    }

    public Configuration GetCurrentConfiguration(IFormat format) {      
      Dictionary<Type, IFormatter> formatters = new Dictionary<Type, IFormatter>();
      foreach ( TabPage page in formatterTabs.TabPages ) {
        if ( page.Text == format.Name ) {
          Control[] controls = page.Controls.Find("GridView", false);
          if (controls.Length == 1) {
            foreach (DataGridViewRow row in ((DataGridView) controls[0]).Rows) {
              if ( row.Cells[1].Value != null &&
                row.Cells[0].Value != null &&
                row.Cells[2].Value != null &&
                (bool)row.Cells[1].Value == true ) {                
                formatters.Add(
                  typeNameTable[(string)row.Cells[0].Value],                  
                  formatterTable[(string)row.Cells[2].Value]);
              }
            }
          }
        }
      }
      return new Configuration(formatters, GetDecomposers());
    }
    
  }

  public class FakeBoolean2XmlFormatter : IFormatter {    

    public Type Type { get { return typeof (Boolean); } }

    public IFormat Format { get { return XmlFormat.Instance; } }

    public object DoFormat(object o) {
      return null;
    }

    public object Parse(object o) {
      return null;
    }
    
  }

  public class FormatterWrapper {
    public IFormatter Formatter;
    public FormatterWrapper(IFormatter f) {
      Formatter = f;
    }
    public override string ToString() {
      return Formatter.GetType().Name;
    }
  }

  public static class TypeFormatter {

    public static string SimpleFullName(this Type type) {
      StringBuilder sb = new StringBuilder();
      SimpleFullName(type, sb);
      return sb.ToString();
    }

    private static void SimpleFullName(Type type, StringBuilder sb) {      
      if (type.IsGenericType) {
        sb.Append(type.Name, 0, type.Name.LastIndexOf('`'));      
        sb.Append("<");
        foreach (Type t in type.GetGenericArguments()) {
          SimpleFullName(t, sb);          
          sb.Append(", ");
        }
        sb.Remove(sb.Length - 2, 2);
        sb.Append(">");
      } else {
        sb.Append(type.Name);
      }
    }

  }

}
